use futures_util::{stream::FuturesUnordered, SinkExt, StreamExt};
use nostr_rs_types::{client::Subscribe, Request, Response};
use parking_lot::RwLock;
use std::{collections::HashMap, sync::Arc};
use tokio::{
    sync::{
        broadcast,
        mpsc::{self, error::SendError},
        oneshot,
    },
    time::{sleep, Duration},
};
use tokio_tungstenite::{
    connect_async, tungstenite::error::Error as TungsteniteError, tungstenite::Message,
};
use url::Url;

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Url: {0}")]
    Url(#[from] url::ParseError),

    #[error("Tungstenite: {0}")]
    Tungstenite(#[from] TungsteniteError),

    #[error("Sync: {0}")]
    Sync(#[from] SendError<Request>),
}

#[derive(Debug)]
pub struct Client {
    pub url: String,

    pub sender: mpsc::Sender<Request>,

    receiver: broadcast::Receiver<(String, Response)>,
    stopper: oneshot::Sender<()>,
}

impl Client {
    pub fn new(url: &str) -> Result<Self, Error> {
        let (sender, receiver) = mpsc::channel(10_000);
        let (receiver, stopper) = Self::spawn(receiver, url)?;

        Ok(Self {
            url: url.to_owned(),
            sender,
            stopper,
            receiver,
        })
    }

    fn spawn(
        mut receiver: mpsc::Receiver<Request>,
        url_str: &str,
    ) -> Result<(broadcast::Receiver<(String, Response)>, oneshot::Sender<()>), Error> {
        let (response_sender, response_receiver) = broadcast::channel(10_000);
        let (stopper_sender, mut stopper_recv) = oneshot::channel();

        let url = url_str.to_owned();
        let url_parsed = Url::parse(&url)?;

        tokio::spawn(async move {
            let mut reconnect = true;
            let mut retries: usize = 0;
            while reconnect && retries <= 10 {
                println!("{}: Connect attempt {}", url, retries);
                retries += 1;
                let mut socket = if let Ok(x) = connect_async(url_parsed.clone()).await {
                    x.0
                } else {
                    println!("{}: Failed to connect", url);
                    sleep(Duration::from_secs(1)).await;
                    continue;
                };

                let request: Request = Subscribe::default().into();
                if let Ok(json) = serde_json::to_string(&request) {
                    if let Err(err) = socket.send(Message::Text(json)).await {
                        println!("{}: Failed to send request {}", url, err);
                        continue;
                    }
                }

                loop {
                    tokio::select! {
                        Ok(()) = &mut stopper_recv => {
                            println!("{}: Breaking client due external signal", url);
                            reconnect = false;
                            break;
                        },
                        Some(msg) = receiver.recv() => {
                            if let Ok(json) = serde_json::to_string(&msg) {
                                if let Err(x) = socket.send(Message::Text(json)).await {
                                    println!("{} :Reconnecting due {}", url, x);
                                    break;

                                }
                            }
                        }
                        Some(Ok(msg)) = socket.next() => {
                            let msg =if let Ok(msg) = msg.into_text() {
                                    msg
                                } else {
                                    continue;
                                };

                            if msg.is_empty() {
                                continue;
                            }

                            let msg: Result<Response, _> = serde_json::from_str(&msg);

                            if let Ok(msg) = msg {
                                if let Err(error) = response_sender.send((url.to_owned(), msg)) {
                                    println!("{}: Reconnecting client because of {}", url, error);
                                    break;
                                }
                            }
                        }
                        else => {
                            println!("{}: else", url);
                        }
                    }
                }
            }

            println!("{}: Disconnected", url);
        });

        Ok((response_receiver, stopper_sender))
    }

    pub fn is_running(&self) -> bool {
        !self.stopper.is_closed()
    }

    pub fn subscribe(&self) -> broadcast::Receiver<(String, Response)> {
        self.receiver.resubscribe()
    }

    pub async fn send(&self, request: Request) -> Result<(), Error> {
        Ok(self.sender.send(request).await?)
    }

    pub async fn stop(self) {
        let _ = self.stopper.send(());
    }
}

#[derive(Debug, Clone)]
pub struct Clients {
    clients: Arc<RwLock<HashMap<String, Client>>>,
    subscriptions: Arc<RwLock<Vec<broadcast::Receiver<(String, Response)>>>>,
}

impl Default for Clients {
    fn default() -> Self {
        Self {
            clients: Arc::new(RwLock::new(HashMap::new())),
            subscriptions: Arc::new(RwLock::new(vec![])),
        }
    }
}

impl Clients {
    pub async fn recv(&self) -> Option<(String, Response)> {
        let mut subscriptions = self
            .subscriptions
            .read()
            .iter()
            .map(|s| s.resubscribe())
            .collect::<Vec<_>>();

        let mut futures = FuturesUnordered::new();

        for sub in subscriptions.iter_mut() {
            futures.push(sub.recv());
        }

        if let Some(Ok(response)) = futures.next().await {
            Some(response)
        } else {
            None
        }
    }
    pub fn try_recv(&self) -> Option<(String, Response)> {
        let mut subscriptions = self.subscriptions.write();
        for sub in subscriptions.iter_mut() {
            if let Ok(msg) = sub.try_recv() {
                return Some(msg);
            }
        }
        None
    }

    pub fn check_active_connections(&self) -> usize {
        let mut clients = self.clients.write();
        let mut subscriptions = self.subscriptions.write();
        let mut to_remove = vec![];
        for (url, client) in clients.iter() {
            if !client.is_running() {
                to_remove.push(url.to_owned());
            }
        }

        for url in to_remove.iter() {
            clients.remove(url);
        }

        subscriptions.retain(|s| s.len() > 0);
        clients.len()
    }

    pub async fn send(&self, request: Request) {
        let senders = self
            .clients
            .read()
            .iter()
            .map(|(_, c)| c.sender.clone())
            .collect::<Vec<mpsc::Sender<_>>>();

        for sender in senders.iter() {
            let _ = sender.send(request.clone()).await;
        }
    }

    pub async fn connect_to(&self, url: &str) -> Result<bool, Error> {
        let mut clients = self.clients.write();
        Ok(if clients.get(url).is_some() {
            false
        } else {
            println!("Connecting to {}", url);
            let client = Client::new(url)?;
            let mut subscriptions = self.subscriptions.write();
            subscriptions.push(client.subscribe());
            clients.insert(url.to_owned(), client);
            true
        })
    }
}
